#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import requests
import json

class ConfParser:
    '''
    Configuration file parser

    :param file_name: Where to store configuration
    '''
    def __init__(self, file_name='/tmp/pci_config.conf'):
        self.file_name = file_name

    def add_entry(self, entry):
        '''Appends a data (as a string) to the file.'''
        try:
            with open(self.file_name, 'a') as file:
                file.write(f'{entry}\n')
        except Exception as e:
            raise Exception(f'Unable to save {entry} to {self.file_name}: {e}')

    def read_entries(self):
        '''Reads all data from the file.'''
        try:
            with open(self.file_name, 'r') as file:
                return file.readlines()
        except FileNotFoundError:
            return []

def conf_rewrite(self, path, data, mode='w+'):
    try:
        with open(path, mode) as f:
            f.write(data)
    except Exception as e:
        self.logger.error(f'Unable to rewrite {path} file!')
        pass

def get_kubevirt_crd(self, controller):
    '''Get current Kubevirt CRD config via controller REST API'''
    self.logger.debug(f'Worker going to get Kubevirt CRDs')

    token = self.get_auth_token_tveapi_controller(controller) # get auth token
    if not token:
        self.logger.error(f'Can not get token from {controller} API')

    url = f'https://{controller}:7373/v1/kubernetes/pci_passthrough'
    headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }

    try:
        response = requests.get(url, headers=headers, verify=False, timeout=10)
        if response.status_code == requests.codes.ok:
            return response.text
    except:
        self.logger.error(f'Unable to get Kubevirt CRDs!')

    return []

def upgrade_kubevirt_crd(self, controller, device):
    '''Worker going to send request for upgrade Kubevirt CRD'''
    self.logger.debug(f'Worker going to send request for upgrade Kubevirt CRD')

    # Suppress multiple API calls
    # config = ConfParser(file_name='/kubernetes/pci_passthrought')
    # kubevirt_crd_accepted = config.read_entries()

    # for pci_device in kubevirt_crd_accepted:
    #     if device == pci_device.split('\n')[0]:
    #         return True

    if self.get_cbox_enabled():
        name = self.get_worker_name()
        if not name:
            self.logger.error('Can not get worker name')
        controller = name

    token = self.get_auth_token_tveapi_controller(controller) # get auth token
    if not token:
        self.logger.error(f'Can not get token from {controller} API')

    url = f'https://{controller}:7373/v1/kubernetes/pci_passthrough'
    headers = {'Content-type': 'application/json', 'X-Auth-Token': token, }
    data = json.dumps({ 
        "operation": "add", 
        "data": [{"externalResourceProvider": False, "pciVendorSelector": device, "resourceName": "vfio/" + device.replace(':', '_')}]
        })

    try:
        # Update Kubevirt CRD via controller's API
        response = requests.post(url, headers=headers, data=data, verify=False, timeout=10)
        if response.status_code == requests.codes.ok:
            self.logger.debug(f'Adding vfio-passthrough {device} to Kubevirt CRD success!')
            # Suppress multiple API calls
            # config.add_entry(device)
            return True
        else:
            self.logger.error(f'Failed to add vfio-passthrough {device} to Kubevirt CRD! Response: {response.text}')

    except:
        self.logger.error(f'Error occurred while trying to add vfio-passthrough {device} to Kubevirt CRD!')

    return False

def attach_device(self, change_set):
    '''We must make sure that PCI device is bound to vfio-pci driver, and not to the native kernel driver.'''
    device = change_set
    device_vendor = device.split(':')[0]
    device_id = device.split(':')[1]
    driver = None

    device_conf = f'/etc/modprobe.d/vfio_{device_vendor}_{device_id}.conf'

    device_pci = '0000:' + self.execute_shell_command(f'lspci -nnk -d {device}').split(' ')[0]

    try:
        if os.path.exists(device_conf):
            os.remove(device_conf)

        config = ConfParser(file_name=device_conf)

        if os.path.islink(f'/sys/bus/pci/devices/{device_pci}/driver'):
            driver = os.readlink(f'/sys/bus/pci/devices/{device_pci}/driver').split('/')[-1]

        # Lets unbind device from his driver, and bind it to vfio-pci
        conf_rewrite(self, f'/sys/bus/pci/devices/{device_pci}/driver_override', 'vfio-pci')
        conf_rewrite(self, '/sys/bus/pci/drivers/vfio-pci/bind', device_pci)

        if driver:
            conf_rewrite(self, f'/sys/bus/pci/drivers/{driver}/unbind', driver)
            config.add_entry(f'blacklist {driver}')

        # Check especially for nVidia
        if driver == 'nvidia' or device_vendor == '10de':
            config.add_entry('blacklist nouveau')
            config.add_entry('blacklist nvidiafb')
            config.add_entry(f'options vfio-pci ids={device} disable_vga=1')
        else:
            config.add_entry(f'options vfio-pci ids={device}')

        os.system('update-initramfs -u')

        upgrade_kubevirt_crd(self, self.get_controller(), device)

        self.logger.info('Calling system_reboot !')
        self._system_reboot()
    except Exception as e:
        self.logger.error(f'Unable to bound PCI device [{device}] to vfio-pci driver - {e}.')

def detach_device(self, change_set):
    device = change_set
    device_vendor = device.split(':')[0]
    device_id = device.split(':')[1]

    device_conf = f'/etc/modprobe.d/vfio_{device_vendor}_{device_id}.conf'

    if os.path.exists(device_conf):
        os.remove(device_conf)
        os.system('update-initramfs -u')

        self.logger.info('Calling system_reboot !')
        self._system_reboot()
    else:
        self.logger.error(f'Unable to unbound PCI device [{device}] to vfio-pci driver. Not found.')
